/* upsd.c - watches ups state files and answers queries 

   Copyright (C) 1999  Russell Kroll <rkroll@exploits.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "upsd.h"
#include "conf.h"

#include "shared-tables.h"
#include "shared-tables-init.h"

#include "netcmds.h"
#include "upsconf.h"

#include <sys/un.h>
#include <sys/socket.h>

#ifdef HAVE_SSL
#include <openssl/err.h>
#include <openssl/ssl.h>
#endif

#include "ctype.h"
#include "ssl.h"
#include "sstate.h"
#include "extstate.h"

#define ACTION_GRANT	1
#define ACTION_DENY	2
#define ACTION_DROP	4

	/* pointers to linked lists */
	upstype	*firstups = NULL;
	acltype *firstacl = NULL;
	acctype *firstacc = NULL;

	ctype	*firstclient = NULL, *currclient = NULL, *udpclient = NULL;

	/* by default TCP uses the same port - use -t to override */
static	int	udpfd, udpport = PORT, upsnum = 0;
static	int	listenfd, tcpport = PORT;

	/* default is to listen on all local interfaces */
static	struct	in_addr	listenaddr;

	/* default 15 seconds before data is marked stale */
	int	maxage = 15;

	/* signal handlers */
	struct sigaction sa;
	sigset_t upsd_sigmask;

	/* pid file */
	char	pidfn[SMALLBUF];

	/* set by SIGHUP */
	int	reload = 0;

	/* preloaded to STATEPATH in main, can be overridden via upsd.conf */
	char	*statepath = NULL;
	const char *progname = NULL;

#ifdef SHUT_RDWR
#define shutdown_how SHUT_RDWR
#else
#define shutdown_how 2
#endif

/* search for a ups given the name */
upstype *findups(const char *upsname)
{
	upstype	*temp;

	temp = firstups;

	/* NULL name -> return first one */
	if (!upsname)
		return firstups;

	/* return first one if a good name isn't given */
	if (!strcmp(upsname, ""))
		return firstups;
	
	while (temp != NULL) {
		if (!strcasecmp(temp->name, upsname))
			return temp;

		temp = temp->next;
	}

	return NULL;
}

/* mark the data stale if this is new, otherwise cleanup any remaining junk */
static void ups_data_stale(upstype *ups)
{
	/* don't complain again if it's already known to be stale */
	if (ups->stale == 1)
		return;

	ups->stale = 1;

	upslogx(LOG_NOTICE, "Data for UPS [%s] is stale - check driver",
		ups->name);
}

/* mark the data ok if this is new, otherwise do nothing */
static void ups_data_ok(upstype *ups)
{
	if (ups->stale == 0)
		return;

	upslogx(LOG_NOTICE, "UPS [%s] data is no longer stale", ups->name);
	ups->stale = 0;
}

/* make sure this UPS is connected and has fresh data */
static void check_ups(upstype *ups)
{
	/* sanity checks */
	if ((!ups) || (!ups->fn))
		return;

	/* throw some warnings if it's not feeding us data any more */
	if (sstate_datastale(ups, maxage))
		ups_data_stale(ups);
	else
		ups_data_ok(ups);

	/* see if we need to (re)connect to the socket */
	if (ups->sock_fd == -1)
		ups->sock_fd = sstate_connect(ups);
}

/* change the configuration of an existing UPS (used during reloads) */
void redefine_ups(char *fn, char *name)
{
	upstype	*temp;

	temp = findups(name);

	if (!temp) {
		upslogx(LOG_ERR, "UPS %s disappeared during reload", name);
		return;
	}

	/* paranoia */
	if (!temp->fn) {
		upslogx(LOG_ERR, "UPS %s had a NULL filename!", name);

		/* let's give it something quick to use later */
		temp->fn = xstrdup("");
	}

	/* when the filename changes, force a reconnect */
	if (strcmp(temp->fn, fn) != 0) {

		upslogx(LOG_NOTICE, "Redefined UPS [%s]", name);

		/* release all data */
		sstate_infofree(temp);
		sstate_cmdfree(temp);
		pconf_finish(&temp->sock_ctx);

		close(temp->sock_fd);
		temp->sock_fd = -1;

		/* now redefine the filename and wrap up */
		free(temp->fn);
		temp->fn = xstrdup(fn);
	}

	/* always set this on reload */
	temp->retain = 1;
}		

/* start monitoring a ups called <name> from a file called <fn> */
void addups(char *fn, char *name)
{
	upstype	*temp, *last;

	temp = last = firstups;

	/* find end of linked list */
	while (temp != NULL) {
		last = temp;

		if (!strcasecmp(temp->name, name)) {
			upslogx(LOG_ERR, "UPS name [%s] is already in use!", 
				name);
			return;
		}

		temp = temp->next;
	}

	/* grab some memory and add the info */
	temp = xcalloc(1, sizeof(upstype));

	temp->fn = xstrdup(fn);

	temp->name = xstrdup(name);
	temp->stale = 1;

	temp->numlogins = 0;
	temp->fsd = 0;
	temp->retain = 1;
	temp->next = NULL;

	temp->sock_fd = -1;
	temp->last_heard = 0;
	temp->last_ping = 0;
	temp->last_connfail = 0;
	temp->inforoot = NULL;

	if (last == NULL)	/* first one */
		firstups = temp;
	else			/* not the first */
		last->next = temp;

	check_ups(temp);
	upsnum++;

	if (temp->stale)
		upslogx(LOG_NOTICE, "Warning: Data for UPS [%s] is stale - check driver", name);
}

/* setup a socket for incoming udp requests */
static void setupudp(void)
{
	struct	sockaddr_in	recv;

	if ((udpfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
		fatal("Can't create socket");

	memset(&recv, '\0', sizeof(recv));
	recv.sin_addr = listenaddr;
	recv.sin_family = AF_INET;
	recv.sin_port = htons(udpport);

	if (bind(udpfd, (struct sockaddr *) &recv, sizeof(recv)) < 0)
		fatal("Can't bind to socket");

	return;
}

/* create a listening socket for tcp connections */
static void setuptcp(void)
{
	struct	sockaddr_in	server;
	int	res, one = 1;

	if ((listenfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		fatal("socket");

	res = setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, (void *) &one, 
		sizeof(one));

	if (res != 0)
		fatal("setsockopt(SO_REUSEADDR)");

	memset(&server, '\0', sizeof(server));
	server.sin_addr = listenaddr;
	server.sin_family = AF_INET;
	server.sin_port = htons(tcpport);

	if (bind(listenfd, (struct sockaddr *) &server, sizeof(server)) == -1)
		fatal("bind");

	if ((res = fcntl(listenfd, F_GETFL, 0)) == -1)
		fatal("fcntl(get)");

	if (fcntl(listenfd, F_SETFL, res | O_NDELAY) == -1)
		fatal("fcntl(set)");

	if (listen(listenfd, 16))
		fatal("listen");

	return;
}

/* decrement the login counter for this ups */
static void declogins(const char *upsname)
{
	upstype	*ups;

	ups = findups(upsname);

	if (ups == NULL) {
		upslogx(LOG_INFO, "Tried to decrement invalid ups name (%s)", upsname);
		return;
	}

	ups->numlogins--;
}

/* remove a client struct from the linked list */
static void delclient(ctype *dclient)
{
	ctype	*tmp, *last;

	if (dclient == NULL)
		return;

	last = NULL;
	tmp = firstclient;

	while (tmp != NULL) {
		if (tmp == dclient) {		/* found it */
			shutdown(dclient->fd, 2);
			close(dclient->fd);

			free(tmp->addr);

			if (tmp->loginups != NULL) {
				declogins(tmp->loginups);
				free(tmp->loginups);
			}

			if (tmp->password != NULL)
				free(tmp->password);

#ifdef HAVE_SSL
			if (tmp->ssl)
				SSL_free(tmp->ssl);
#endif

			if (last == NULL)	/* deleting first entry */
				firstclient = tmp->next;
			else
				last->next = tmp->next;

			free(tmp);
			currclient = NULL;
			return;
		}
		last = tmp;
		tmp = tmp->next;
	}

	/* not found?! */
	upslogx(LOG_WARNING, "Tried to delete client struct that doesn't exist!");

	return;
}

/* send the buffer <sendbuf> of length <sendlen> to host <dest> */
int sendback(const struct sockaddr_in *dest, const char *fmt, ...)
{
#define NUT_NET_ANSWER_MAX SMALLBUF
	int	res, destfd, len;
	char ans[NUT_NET_ANSWER_MAX+1];
	va_list ap;

	/* answer_tcp -> checkaccess (deny) might call this */
	if (!currclient)
		return 0;	/* failed */

	if (currclient == udpclient)
		destfd = udpfd;
	else
		destfd = currclient->fd;

	va_start(ap, fmt);
	vsnprintf(ans, sizeof(ans), fmt, ap);
	va_end(ap);

	len = strlen(ans);

	upsdebugx(2, "sendto: [destfd=%d] [len=%d] [%s]", destfd, len, ans);

	if (currclient->ssl)
		res = ssl_write(currclient, ans, len);
	else
		res = sendto(destfd, ans, len, 0, (struct sockaddr *) dest,
	              sizeof(struct sockaddr_in));

	if (len != res) {
		upslog(LOG_NOTICE, "sendto() failed for %s", currclient->addr);
		delclient(currclient);
		return 0;	/* failed */
	}

	return 1;	/* OK */
}

/* see if <addr> matches the acl <aclname> */
int checkacl(const char *aclname, const struct sockaddr_in *addr)
{
	acltype	*tmp;
	int	aclchk, addrchk;

	tmp = firstacl;
	while (tmp != NULL) {
		if (!strcmp(tmp->name, aclname)) {
			aclchk = tmp->addr & tmp->mask;
			addrchk = ntohl(addr->sin_addr.s_addr) & tmp->mask;

			if (aclchk == addrchk) 
				return 1;	/* match */
		}

		tmp = tmp->next;
	}

	return 0;	/* not found */
}

/* this is only used for simple stuff: MONITOR and BASE */
static int checkaccess(const struct sockaddr_in *ip, int level)
{
	acctype	*tmp;
	int	ret;

	/* check for insanity */
	if (level > LEVEL_MONITOR) {
		upslogx(LOG_ERR, "checkaccess called with level=%d", level);
		return 0;	/* failed */
	}

	tmp = firstacc;

	while (tmp != NULL) {
		ret = checkacl(tmp->aclname, ip);

		upsdebugx(3, "checkacl: %s: match %d, level %d (%d)", 
			tmp->aclname, ret, tmp->level, level);

		/* if ip matches and access line provides the right level.. */

		if ((ret == 1) && (tmp->level >= level)) {

			if (tmp->action == ACTION_GRANT)
				return 1;	/* allowed */

			if (tmp->action == ACTION_DENY) {
				upsdebugx(1, "access denied");
				sendback(ip, "ERR ACCESS-DENIED\n");
			}

			/* anything else is treated as a DROP - no response */
			return 0;

		} 	/* if (ip matches) && (level is provided) */

		/* otherwise ip didn't match or the level was inadequate */

		tmp = tmp->next;
	}

	/* default = drop */
	return 0;
}

/* increment the login counter for this ups */
static void inclogins(const char *upsname)
{
	upstype	*ups;

	ups = findups(upsname);

	if (ups == NULL) {
		upslogx(LOG_INFO, "Tried to increment invalid ups name");
		return;
	}

	ups->numlogins++;
}

/* disconnect anyone logged into this UPS */
void kick_login_clients(char *upsname)      
{
	ctype   *tmp, *next;

	tmp = firstclient;

	while (tmp) {
		next = tmp->next;

		/* if it's not logged in, don't check it */
		if (!tmp->loginups) {
			tmp = next;
			continue;
		}

		if (!strcmp(tmp->loginups, upsname)) {
			upslogx(LOG_INFO, "Kicking client %s (was on UPS [%s])\n", 
				tmp->addr, upsname);
			delclient(tmp);
		}
                
		tmp = next;
	}
}

/* parse varname[@upsname] into separate variables */
static upstype *parsearg(const char *arg, char **varname)
{
	char	*argtmp, *upsname;
	upstype	*ups;

	argtmp = xstrdup(arg);

	upsname = strchr(argtmp, '@');
	if (upsname) {
		*upsname++ = '\0';
		ups = findups(upsname);
	} else {
		ups = firstups;
	}

	/* varname may be a subset of argtmp, so copy it and free argtmp */
	*varname = xstrdup(argtmp);
	free(argtmp);

	return ups;
}

/* make sure a UPS is sane - connected, with fresh data */
static int ups_available(upstype *ups, struct sockaddr_in *dest)
{
	if (ups->sock_fd == -1) {
		sendback(dest, "ERR DRIVER-NOT-CONNECTED\n");
		return 0;
	}

	if (sstate_datastale(ups, maxage)) {
		sendback(dest, "ERR DATA-STALE\n");
		return 0;
	}

	/* must be OK */
	return 1;
}

/* handler for "REQ" - send a reply */
void do_sendans(struct sockaddr_in *dest, char *varin, char *rest)
{
	int	i, type = 0;
	char	ans[SMALLBUF], *varname;
	upstype	*ups;

	char	tempvn[SMALLBUF];
	const	char	*tempval;

	if (varin == NULL) {
		sendback(dest, "ERR MISSING-ARGUMENT\n");
		return;
	}

	ups = parsearg(varin, &varname);

	if (ups == NULL) {
		sendback(dest, "ERR UNKNOWN-UPS\n", varin);
		free(varname);
		return;
	}

	/* special case: numlogins is an internal value */
	if (!strcasecmp(varname, "NUMLOGINS")) {
		sendback(dest, "ANS %s %d\n", varname, ups->numlogins);
		free(varname);
		return;
	}

	/* make sure this thing is still with us */
	if (!ups_available(ups, dest)) {
		free(varname);
		return;
	}

	for (i = 0; netvars[i].name != NULL; i++)
		if (!strcasecmp(netvars[i].name, varname))
			type = netvars[i].type;

	free(varname);
	
	/* type wasn't resolved in the netvars[] array */
	if (type == 0) {
		sendback(dest, "ERR VAR-UNKNOWN\n", varin);
		return;
	}

	/* make sure this thing is still with us */
	if (!ups_available(ups, dest))
		return;

	/* mapping old numerics into new text strings for now */
	snprintf(tempvn, sizeof(tempvn), "%04X", type);

	tempval = sstate_getinfo(ups, tempvn);

	if (!tempval) {
		sendback(dest, "ERR VAR-NOT-SUPPORTED\n");
		return;
	}

	/* special case - make FSD show up in STATUS */
	if ((type == INFO_STATUS) && (ups->fsd == 1)) 
		snprintf(ans, sizeof(ans), "ANS %s FSD %s", varin, tempval);
	else
		snprintf(ans, sizeof(ans), "ANS %s %s",	varin, tempval);

	sendback(dest, "%s\n", ans);
}

/* handler for "LISTVARS" */
void do_listvars(struct sockaddr_in *dest, char *arg, char *rest)
{
	char	ans[SMALLBUF];
	upstype *ups;

	ups = findups(arg);

	if (ups == NULL) {
		sendback(dest, "ERR UNKNOWN-UPS\n");
		return;
	}

	/* make sure this thing is still with us */
	if (!ups_available(ups, dest))
		return;

	memset(ans, '\0', sizeof(ans));

	/* insert "@upsname" if explicitly specified in request */
	if ((arg) && (strcmp(arg, "") != 0))
		snprintf(ans, sizeof(ans), " @%s", arg);

	sstate_makeinfolist(ups, ans, sizeof(ans));

	sendback(dest, "VARS%s\n", ans);
}

/* handler for "VER" */
void do_sendver(struct sockaddr_in *dest, char *arg, char *rest)
{
	sendback(dest, "Network UPS Tools upsd %s - http://www.exploits.org/nut/\n", UPS_VERSION);
}

/* handler for "HELP" */
void do_sendhelp(struct sockaddr_in *dest, char *arg, char *rest)
{
	char	ans[SMALLBUF];
	int	i;

	strlcpy(ans, "Commands:", sizeof(ans));

	for (i = 0; netcmds[i].name != NULL; i++)
		snprintfcat(ans, sizeof(ans), " %s", netcmds[i].name);

	sendback(dest, "%s\n", ans);
}

/* handler for "LOGOUT" */
void do_logout(struct sockaddr_in *dest, char *arg, char *rest)
{
	upslogx(LOG_INFO, "Client on %s logged out", currclient->addr);

	/* check return since sendback can call delclient itself */
	if (sendback(dest, "OK Goodbye\n"))
		delclient(currclient);
}		

/* handler for "LOGIN" */
static void do_login(struct sockaddr_in *dest, char *arg, char *rest)
{
	int	found = 0;
	upstype *ups;

	if (currclient->loginups != NULL) {
		upslogx(LOG_INFO, "Client on %s tried to login twice", currclient->addr);
		sendback(dest, "%s", "ERR ALREADY-LOGGED-IN\n");
		return;
	}

	/* "LOGIN" only - no UPS name given as argument */
	if ((arg == NULL) || (strlen(arg) == 0)) {

		/* no UPSes defined? */
		if (!firstups) {
			sendback(dest, "ERR UNKNOWN-UPS\n");
			return;
		}

		/* give it the name of the first one in the .conf */
		currclient->loginups = xstrdup(firstups->name);

		/* but warn them that this is going away some day */
		upslogx(LOG_WARNING, "Warning: LOGIN relied on the default UPS");
	}
	else {
		/* see if <arg> is a valid UPS name */

		for (ups = firstups; ups != NULL; ups = ups->next) {
			if (!strcasecmp(ups->name, arg))
				found = 1;
		}

		if (!found) {
			sendback(dest, "ERR UNKNOWN-UPS\n");
			return;
		}

		currclient->loginups = xstrdup(arg);
	}

	inclogins(currclient->loginups);

	upslogx(LOG_INFO, "Client %s@%s logged into UPS [%s]%s",
		currclient->username, currclient->addr, 
		currclient->loginups, 
		currclient->ssl ? " (SSL)" : "");

	sendback(dest, "OK\n");
}

/* handler for "PASSWORD" */
static void do_password(struct sockaddr_in *dest, char *arg, char *rest)
{
	if ((arg == NULL) || (strlen(arg) == 0)) {
		upslogx(LOG_INFO, "Client on %s sent a NULL password!", currclient->addr);
		sendback(dest, "ERR INVALID-PASSWORD\n");
		return;	
	}

	if (currclient->password != NULL) {
		upslogx(LOG_INFO, "Client on %s tried to set a password twice", currclient->addr);
		sendback(dest, "ERR ALREADY-SET-PASSWORD\n");
		return;
	}

	currclient->password = xstrdup(arg);

	sendback(dest, "OK\n");
}

/* handler for "LISTRW" */
static void do_listrw(struct sockaddr_in *dest, char *arg, char *rest)
{
	char	ans[SMALLBUF];
	upstype *ups;

	ups = findups(arg);

	if (ups == NULL) {
		sendback(dest, "ERR UNKNOWN-UPS\n");
		return;
	}

	/* make sure this thing is still with us */
	if (!ups_available(ups, dest))
		return;

	ans[0] = '\0';

	/* insert "@upsname" if explicitly specified in request */
	if ((arg) && (strcmp(arg, "") != 0)) {
		snprintfcat(ans, sizeof(ans), " @%s", arg);
	}		

	sstate_makerwlist(ups, ans, sizeof(ans));

	sendback(dest, "RW%s\n", ans);
}

/* handler for "VARTYPE" */
static void do_vartype(struct sockaddr_in *dest, char *arg, char *rest)
{
	char	*varname;
	int	i, type;
	upstype *ups;

	char	tempvn[SMALLBUF];
	int	flags, ecount;
	const	struct	enum_t	*etmp;

	if (arg == NULL) {
		sendback(dest, "ERR MISSING-ARGUMENT\n");
		return;
	}

	ups = parsearg(arg, &varname);

	if (ups == NULL) {
		sendback(dest, "ERR UNKNOWN-UPS\n");
		free(varname);
		return;
	}

	type = 0;

	/* find the variable type for the name */
	for (i = 0; netvars[i].name != NULL; i++)
		if (!strcasecmp(netvars[i].name, varname))
			type = netvars[i].type;

	if (type == 0) {	/* didn't find it */
		sendback(dest, "ERR UNKNOWN-TYPE\n");
		free(varname);
		return;
	}

	/* no longer needed */
	free(varname);

	/* mapping old numerics into new text strings for now */
	snprintf(tempvn, sizeof(tempvn), "%04X", type);

	flags = sstate_getflags(ups, tempvn);

	/* see if this variable even exists */
	if (flags == -1) {
		sendback(dest, "ERR VAR-NOT-SUPPORTED\n");
		return;
	}

	/* maybe it's a string */
	if (flags & ST_FLAG_STRING) {
		sendback(dest, "TYPE STRING %d\n", sstate_getaux(ups, tempvn));
		return;
	}

	/* maybe it's an enum */
	etmp = sstate_getenumlist(ups, tempvn);

	ecount = 0;
	while (etmp) {
		ecount++;
		etmp = etmp->next;
	}

	if (ecount) {
		sendback(dest, "TYPE ENUM %d\n", ecount);
		return;
	}

	/* unknown */
	sendback(dest, "ERR UNKNOWN-TYPE\n");
}

/* TODO: clean this up ... lots of identical code in here */

/* handler for "VARDESC" */
static void do_vardesc(struct sockaddr_in *dest, char *arg, char *rest)
{
	char	ans[SMALLBUF];
	int	i;

	/* stock reply */
	snprintf(ans, sizeof(ans), "ERR VAR-NOT-SUPPORTED\n");

	/* find the variable type for the name */
	for (i = 0; netvars[i].name != NULL; i++)
		if (!strcasecmp(netvars[i].name, arg))
			snprintf(ans, sizeof(ans), "DESC \"%s\"\n", 
				netvars[i].desc);

	sendback(dest, "%s", ans);
}

/* handler for "ENUM" */
static void do_enum(struct sockaddr_in *dest, char *arg, char *rest)
{
	char	*varname;
	int	i, type;
	upstype *ups;

	char	tempvn[SMALLBUF];
	const	char	*tempval;
	const	struct	enum_t	*etmp;

	if (arg == NULL) {
		sendback(dest, "ERR MISSING-ARGUMENT\n");
		return;
	}

	ups = parsearg(arg, &varname);

	if (ups == NULL) {
		sendback(dest, "ERR UNKNOWN-UPS\n");
		free(varname);
		return;
	}

	type = 0;

	/* find the variable type for the name */
	for (i = 0; netvars[i].name != NULL; i++)
		if (!strcasecmp(netvars[i].name, varname))
			type = netvars[i].type;

	if (type == 0) {	/* didn't find it */
		sendback(dest, "ERR UNKNOWN-TYPE\n");
		free(varname);
		return;
	}

	/* mapping old numerics into new text strings for now */
	snprintf(tempvn, sizeof(tempvn), "%04X", type);

	/* see if this variable even exists first */
	tempval = sstate_getinfo(ups, tempvn);

	if (!tempval) {
		sendback(dest, "ERR VAR-NOT-SUPPORTED\n");
		free(varname);
		return;
	}

	/* bail out here if the write fails */
	if (!sendback(dest, "ENUM %s\n", varname)) {
		free(varname);
		return;
	}

	/* don't need this any more */
	free(varname);

	etmp = sstate_getenumlist(ups, tempvn);

	while (etmp) {
		char	resp[SMALLBUF];

		if (!strcmp(etmp->val, tempval))
			snprintf(resp, sizeof(resp), "OPTION \"%s\" SELECTED",
				etmp->val);
		else
			snprintf(resp, sizeof(resp), "OPTION \"%s\"",
				etmp->val);

		if (!sendback(dest, "%s\n", resp))
			return;

		etmp = etmp->next;
	}

	sendback(dest, "END\n");
}

/* handler for "SET" */
static void do_set(struct sockaddr_in *dest, char *arg, char *rest)
{
	char	*newval, *varname, *ptr;
	upstype *ups;
	int	i, type, found;

	char	tempvn[SMALLBUF], cmd[SMALLBUF], esc[SMALLBUF];
	const	char	*tempval;
	const	struct	enum_t	*etmp;

	if ((arg == NULL) || (rest == NULL)) {
		sendback(dest, "ERR MISSING-ARGUMENT\n");
		return;
	}

	ups = parsearg(arg, &varname);

	if (ups == NULL) {
		sendback(dest, "ERR UNKNOWN-UPS\n");
		free(varname);
		return;
	}

	type = 0;
	/* make sure 'varname' is part of the protocol as we know it */
	for (i = 0; netvars[i].name != NULL; i++)
		if (!strcasecmp(netvars[i].name, varname))
			type = netvars[i].type;
	
	/* type wasn't resolved in the netvars[] array */
	if (type == 0) {
		sendback(dest, "ERR UNKNOWN-VAR\n");
		free(varname);
		return;
	}

	/* make sure this thing is still with us */
	if (!ups_available(ups, dest)) {
		free(varname);
		return;
	}

	/* mapping old numerics into new text strings for now */
	snprintf(tempvn, sizeof(tempvn), "%04X", type);

	/* see if this UPS supports that variable */
	tempval = sstate_getinfo(ups, tempvn);

	if (!tempval) {
		sendback(dest, "ERR VAR-NOT-SUPPORTED\n");
		free(varname);
		return;
	}

	/* make sure this variable is writable (RW) */
	if ((sstate_getflags(ups, tempvn) & ST_FLAG_RW) == 0) {
		sendback(dest, "ERR READONLY\n");
		free(varname);
		return;
	}

	/* clean up rest into something usable */
	newval = xstrdup(rest);

	ptr = strchr(newval, 13);
	if (ptr)
		*ptr = '\0';	/* strip trailing CR */

	ptr = strchr(newval, 10);
	if (ptr)
		*ptr = '\0';	/* strip trailing LF */

	/* finally, see if the new value is allowed for this variable */

	/* check the length if this is for a string */
	if (sstate_getflags(ups, tempvn) & ST_FLAG_STRING) {
		if (strlen(newval) > sstate_getaux(ups, tempvn)) {
			sendback(dest, "ERR TOO-LONG\n");
			free(newval);
			free(varname);
			return;
		}
	}

	/* see if it's enumerated */
	etmp = sstate_getenumlist(ups, tempvn);

	if (etmp) {
		found = 0;

		/* make sure this value is supported */

		while (etmp) {
			if (!strcmp(etmp->val, newval)) {
				found = 1;
				break;
			}

			etmp = etmp->next;
		}

		if (!found) {
			sendback(dest, "ERR INVALID-VALUE\n");
			free(newval);
			free(varname);
			return;
		}
	}
	
	upslogx(LOG_INFO, "Set variable: %s@%s set %s on %s to %s",
	       currclient->username, currclient->addr, varname, ups->name, 
	       newval);

	snprintf(cmd, sizeof(cmd), "SET %s \"%s\"\n",
		tempvn, pconf_encode(newval, esc, sizeof(esc)));

	if (sstate_sendline(ups, cmd)) {
		sendback(dest, "OK\n");
	} else {
		upslogx(LOG_INFO, "Set command send failed");
		sendback(dest, "ERR SET-FAILED\n");
	}

	free(newval);
	free(varname);
}

/* handler for "INSTCMD" */
static void do_instcmd(struct sockaddr_in *dest, char *arg, char *rest)
{
	char	*cmdname;
	upstype *ups;
	int	i, cmd, found;

	char	tempcn[SMALLBUF], sockcmd[SMALLBUF];
	const	struct	cmdlist_t	*ctmp;

	if (!arg) {
		sendback(dest, "ERR MISSING-ARGUMENT\n");
		return;
	}		

	ups = parsearg(arg, &cmdname);

	if (ups == NULL) {
		sendback(dest, "ERR UNKNOWN-UPS\n");
		free(cmdname);
		return;
	}

	cmd = 0;

	/* walk through instcmds[], strcmp */
	for (i = 0; instcmds[i].name != NULL; i++)
		if (!strcasecmp(instcmds[i].name, cmdname))
			cmd = instcmds[i].cmd;

	if (cmd == 0) {
		sendback(dest, "ERR UNKNOWN-INSTCMD\n");
		free(cmdname);
		return;
	}

	/* now make sure the UPS can actually DO this command */

	/* mapping old numerics into new text strings for now */
	snprintf(tempcn, sizeof(tempcn), "%04X", cmd);

	ctmp = sstate_getcmdlist(ups);

	found = 0;
	while (ctmp) {
		if (!strcmp(ctmp->name, tempcn)) {
			found = 1;
			break;
		}

		ctmp = ctmp->next;
	}

	if (!found) {
		sendback(dest, "ERR CMD-NOT-SUPPORTED\n");
		free(cmdname);
		return;
	}

	/* see if this user is allowed to do this command */
	if (!user_checkinstcmd(dest, currclient->username, 
		currclient->password, cmd)) {
			sendback(dest, "ERR ACCESS-DENIED\n");
			free(cmdname);
			return;
		}

	upslogx(LOG_INFO, "Instant command: %s@%s did %s on %s",
	       currclient->username, currclient->addr, cmdname, 
	       ups->name);

	snprintf(sockcmd, sizeof(sockcmd), "INSTCMD %s\n", tempcn);

	if (sstate_sendline(ups, sockcmd)) {
		sendback(dest, "OK\n");
	} else {
		upslogx(LOG_INFO, "Set command send failed");
		sendback(dest, "ERR INSTCMD-FAILED\n");
	}

	free(cmdname);
}

/* handler for "LISTINSTCMD" */
static void do_listinstcmd(struct sockaddr_in *dest, char *arg, char *rest)
{
	char	ans[SMALLBUF];
	upstype *ups;

	ups = findups(arg);

	if (ups == NULL) {
		sendback(dest, "ERR UNKNOWN-UPS\n");
		return;
	}

	/* make sure this thing is still with us */
	if (!ups_available(ups, dest))
		return;

	ans[0] = '\0';

	/* insert "@upsname" if explicitly specified in request */
	if ((arg) && (strcmp(arg, "") != 0))
		snprintfcat(ans, sizeof(ans), " @%s", arg); 

	sstate_makeinstcmdlist(ups, ans, sizeof(ans));

	sendback(dest, "INSTCMDS%s\n", ans);
}

/* handler for "INSTCMDDESC" */
static void do_instcmddesc(struct sockaddr_in *dest, char *arg, char *rest)
{
	char	ans[SMALLBUF];
	int	i;

	if (arg == NULL) {
		sendback(dest, "ERR MISSING-ARGUMENT\n");
		return;
	}

	/* stock reply */
	snprintf(ans, sizeof(ans), "ERR CMD-NOT-SUPPORTED\n");

	/* find the variable type for the name */
	for (i = 0; instcmds[i].name != NULL; i++)
		if (!strcasecmp(instcmds[i].name, arg))
			snprintf(ans, sizeof(ans), "DESC \"%s\"\n", 
				instcmds[i].desc);

	sendback(dest, "%s", ans);
}

/* handler for "FSD" */
static void do_fsd(struct sockaddr_in *dest, char *arg, char *rest)
{
	upstype *ups;

	ups = findups(arg);

	if (ups == NULL) {
		sendback(dest, "ERR UNKNOWN-UPS\n");
		return;
	}

	upslogx(LOG_INFO, "Setting FSD on UPS [%s]", ups->name);

	ups->fsd = 1;
	sendback(dest, "OK FSD-SET\n");
}

/* handler for "MASTER" */
static void do_master(struct sockaddr_in *dest, char *arg, char *rest)
{
	upstype *ups;

	ups = findups(arg);

	if (ups == NULL) {
		sendback(dest, "ERR UNKNOWN-UPS\n");
		return;
	}

	/* really quite simple - this is just an access level check */
	sendback(dest, "OK MASTER-GRANTED\n");
}

/* handler for "USERNAME" */
static void do_username(struct sockaddr_in *dest, char *arg, char *rest)
{
	if ((arg == NULL) || (strlen(arg) ==0)) {
		upslogx(LOG_INFO, "Client on %s sent a NULL username!",
		       currclient->addr);
		sendback(dest, "ERR INVALID-USERNAME\n");
		return;
	}

	if (currclient->username != NULL) {
		upslogx(LOG_INFO, "Client on %s tried to set a username twice",
		       currclient->addr);
		sendback(dest, "ERR ALREADY-SET-USERNAME\n");
		return;
	}

	currclient->username = xstrdup(arg);

	sendback(dest, "OK\n");
}

static int check_flags(struct sockaddr_in *dest, int flags, const char *action)
{
	if (flags & FLAG_TCP) {
		if (currclient == udpclient) {
			sendback(dest, "ERR Impossible in UDP mode\n");
			return 0;
		}
	}

	if (flags & FLAG_USER) {
		if (!currclient->username) {
			sendback(dest, "ERR USERNAME-REQUIRED\n");
			return 0;
		}

		if (!currclient->password) {
			sendback(dest, "ERR PASSWORD-REQUIRED\n");
			return 0;
		}

		/* instcmd is a special case - checked in do_instcmd */
		if (flags & FLAG_CMD)
			return 1;

		/* now check this specific action */

		if (!user_checkaction(dest, currclient->username,
			currclient->password, action)) {
			sendback(dest, "ERR ACCESS-DENIED\n");
			return 0;
		}
	}

	/* seems good so far... */
	return 1;
}

/* parse requests from the network */
static void parse(int len, char *buf, struct sockaddr_in *from)
{
	int	i;
	char	*cmd, *arg2, *rest, *ptr;

	/* check for base access if this is a UDP client */
	if ((currclient == udpclient) && (!checkaccess(from, LEVEL_BASE)))
		return;

	arg2 = rest = NULL;

	/* kill any embedded cr/lf chars */
	for (i = 0; i < len; i++)
		if ((buf[i] == 13) || (buf[i] == 10))
			buf[i] = '\0';

	/* parse from buf: <cmd> [<arg2> [<rest>]] */

	cmd = buf;
	ptr = strchr(cmd, ' ');
	if (ptr) {
		*ptr++ = '\0';
		arg2 = ptr;
		ptr = strchr(arg2, ' ');
		if (ptr) {
			*ptr++ = '\0';
			rest = ptr;
		}
	}

	for (i = 0; netcmds[i].name != NULL; i++) {
		if (!strcasecmp(netcmds[i].name, cmd)) {

			/* see if the flags are satisfied */
			if (!check_flags(from, netcmds[i].flags, cmd))
				return;

			if (!checkaccess(from, netcmds[i].level))
				return;

			netcmds[i].func(from, arg2, rest);
			return;
		}
	}

	upsdebugx(4, "Unknown command: %s", cmd);
	sendback(from, "ERR UNKNOWN-COMMAND\n");
}

/* scan the list of UPSes for sanity */
void check_every_ups(void)
{
	upstype *ups;

	ups = firstups;

	while (ups != NULL) {
		check_ups(ups);
		ups = ups->next;
	}
}

/* add the acl to the linked list */
void addacl (char *aclname, char *ipblock)
{
	acltype	*tmp, *last;
	char	*addr, *mask;

	/* are we sane? */
	if ((!aclname) || (!ipblock))
		return;

	/* ipblock must be in the form <addr>/<mask>	*/

	mask = strchr (ipblock, '/');

	/* 206.253.95.200/32 : VALID			*/
	/* 206.253.95.200/255.255.255.255 : VALID	*/
	/* 206.253.95.200 : INVALID			*/

	if (!mask) 	/* no slash = broken acl declaration */
		return;

	*mask++ = '\0';
	addr = ipblock;

	tmp = last = firstacl;

	while (tmp != NULL) {	/* find end */
		last = tmp;
		tmp = tmp->next;
	}

	tmp = xmalloc (sizeof (acltype));
	tmp->name = xstrdup (aclname);
	tmp->addr = ntohl(inet_addr (addr));
	tmp->next = NULL;

	if (strstr(mask, ".") == NULL) { /* must be a /nn CIDR type block */
		if (atoi(mask) != 32)
			tmp->mask = ((unsigned int) ((1 << atoi(mask)) - 1) << 
			            (32 - atoi(mask)));
		else
			tmp->mask = 0xffffffff;	/* avoid overflow from 2^32 */
	}
	else
		tmp->mask = ntohl(inet_addr (mask));

	if (last == NULL)	/* first */
		firstacl = tmp;
	else
		last->next = tmp;
}

/* add to the access linked list */
void addaccess(const char *action, char *level, char *aclname, char *password)
{
	acctype	*tmp, *last;

	/* more sanity checking (null password is OK) */
	if ((!action) || (!level) || (!aclname))
		return;

	tmp = last = firstacc;

	while (tmp != NULL) {	/* find end */
		last = tmp;
		tmp = tmp->next;
	}

	tmp = xmalloc (sizeof(acctype));
	tmp->action = 0;
	tmp->level = 0;

	if (!strcasecmp(action, "grant"))
		tmp->action = ACTION_GRANT;
	if (!strcasecmp(action, "deny"))
		tmp->action = ACTION_DENY;
	if (!strcasecmp(action, "drop"))
		tmp->action = ACTION_DROP;

	if (!strcasecmp(level, "base"))
		tmp->level = LEVEL_BASE;
	if (!strcasecmp(level, "monitor"))
		tmp->level = LEVEL_MONITOR;

	/* deprecated in 1.2, removed in 1.3 */
	if ((!strcasecmp(level, "login")) ||
	   (!strcasecmp(level, "master"))) {

		upslogx(LOG_WARNING, "LOGIN and MASTER no longer supported "
			" in upsd.conf - switch to upsd.users");

		/* give them something somewhat useful */
		tmp->level = LEVEL_MONITOR;
	}

	/* deprecated in 1.0, removed in 1.1 */
	if (!strcasecmp(level, "manager")) {
		upslogx(LOG_WARNING, "ACCESS type manager no longer supported -"
			" switch to upsd.users");

		/* but don't leave them totally out in the cold */
		tmp->level = LEVEL_MONITOR;
	}
		
	if (!strcasecmp(level, "all"))
		tmp->level = LEVEL_ALL;

	tmp->aclname = xstrdup(aclname);
	tmp->next = NULL;

	if ((password == NULL) || (strlen(password) == 0))
		tmp->password = NULL;
	else
		tmp->password = xstrdup(password);

	if (last == NULL)	/* first */
		firstacc = tmp;
	else
		last->next = tmp;	
}

/* answer incoming tcp connections */
static void answertcp(void)
{
	int	acc;
	struct	sockaddr_in client;
	ctype	*tmp, *last;
	int	clen;

	clen = sizeof(client);
	acc = accept(listenfd, (struct sockaddr *) &client, &clen);

	if (acc < 0)
		return;

	if (!checkaccess(&client, LEVEL_BASE)) {
		upslogx(LOG_NOTICE, "Rejecting TCP connection from %s", 
			inet_ntoa(client.sin_addr));
		shutdown(acc, shutdown_how);
		close(acc);
		return;
	}

	upslogx(LOG_INFO, "Connection from %s", inet_ntoa(client.sin_addr));

	last = tmp = firstclient;

	while (tmp != NULL) {
		last = tmp;
		tmp = tmp->next;
	}

	tmp = xmalloc(sizeof(ctype));

	tmp->addr = xstrdup(inet_ntoa(client.sin_addr));
	tmp->fd = acc;
	memcpy(&tmp->sock, &client, sizeof(struct sockaddr_in));

	tmp->rqpos = 0;
	memset(tmp->rq, '\0', sizeof(tmp->rq));

	tmp->loginups = NULL;		/* for upsmon */
	tmp->username = NULL;
	tmp->password = NULL;

	tmp->ssl = NULL;
	tmp->ssl_connected = 0;

	tmp->next = NULL;

	if (last == NULL)
 		firstclient = tmp;
	else
		last->next = tmp;
}

/* read tcp messages and handle them */
static void readtcp(ctype *client)
{
	char	buf[SMALLBUF];
	int	i, ret;

	memset(buf, '\0', sizeof(buf));

	if (client->ssl)
		ret = ssl_read(client, buf, sizeof(buf));
	else
		ret = read(client->fd, buf, sizeof(buf));

	if (ret < 1) {
		upslogx(LOG_INFO, "Host %s disconnected (read failure)",
			client->addr);

		delclient(client);
		return;
	}

	/* if an overflow will happen, then clear the queue */
	if ((ret + client->rqpos) >= sizeof(client->rq)) {
		memset(client->rq, '\0', sizeof(client->rq));
		client->rqpos = 0;
	}

	/* fragment handling code */

	for (i = 0; i < ret; i++) {
		/* add to the receive queue one by one */
		client->rq[client->rqpos++] = buf[i];

		/* parse on linefeed ('blah blahCRLF') */
		if (buf[i] == 10) {	/* LF */
			currclient = client;	/* enable tcp replies */
			parse(client->rqpos, client->rq, &client->sock);

			if (currclient == NULL) /* connection was closed */
				break;

			currclient = NULL;	/* disable */
			
			/* reset queue */
			client->rqpos = 0;
			memset(client->rq, '\0', sizeof(client->rq));
		}
	}

	return;
}

/* service requests and check on new data */
static void mainloop(void)
{
	fd_set	rfds;
	struct	timeval	tv;
	struct	sockaddr_in from;
	int	res, maxfd;
	char	buf[SMALLBUF];
	ctype	*tmpcli, *tmpnext;
	upstype	*utmp, *unext;
	int	fromlen;

	if (reload) {
		reload_config();
		reload = 0;
	}

	FD_ZERO(&rfds);
	FD_SET(udpfd, &rfds);
	FD_SET(listenfd, &rfds);
	
	tv.tv_sec = 2;
	tv.tv_usec = 0;

	maxfd = (udpfd > listenfd) ? udpfd : listenfd;

	/* scan through clients and add to FD_SET */
	for (tmpcli = firstclient; tmpcli != NULL; tmpcli = tmpcli->next) {
		FD_SET(tmpcli->fd, &rfds);
		if (tmpcli->fd > maxfd)
			maxfd = tmpcli->fd;
	}

	/* also add new driver sockets */
	for (utmp = firstups; utmp != NULL; utmp = utmp->next) {
		if (utmp->sock_fd != -1) {
			FD_SET(utmp->sock_fd, &rfds);

			if (utmp->sock_fd > maxfd)
				maxfd = utmp->sock_fd;
		}
	}
	
	res = select(maxfd + 1, &rfds, NULL, NULL, &tv);

	if (res > 0) {
		if (FD_ISSET (udpfd, &rfds)) {
			fromlen = sizeof(from);
			memset(buf, '\0', sizeof(buf));
			res = recvfrom(udpfd, buf, sizeof(buf), 0, 
				(struct sockaddr *) &from, &fromlen);

			if (res > 0) {

				/* dummy setting, as opposed to NULL */
				currclient = udpclient;

				parse(res, buf, &from);

				currclient = NULL;
			}
		}

		if (FD_ISSET(listenfd, &rfds))
			answertcp();

		/* scan clients for activity */

		tmpcli = firstclient;

		while (tmpcli != NULL) {

			/* preserve for later since delclient may run */
			tmpnext = tmpcli->next;

			if (FD_ISSET(tmpcli->fd, &rfds))
				readtcp(tmpcli);

			tmpcli = tmpnext;
		}

		/* now scan ups sockets for activity */
		utmp = firstups;

		while (utmp) {
			unext = utmp->next;

			if (utmp->sock_fd != -1)
				if (FD_ISSET(utmp->sock_fd, &rfds))
					sstate_sock_read(utmp);

			utmp = unext;
		}
	}

	check_every_ups();
}

static void help(const char *progname) 
{
	printf("Network server for UPS data.\n\n");
	printf("usage: %s [OPTIONS]\n", progname);

	printf("\n");
	printf("  -c <command>	send <command> via signal to background process\n");
	printf("		commands:\n");
	printf("		 - reload: reread configuration files\n");
	printf("		 - stop: stop process and exit\n");
	printf("  -D		raise debugging level\n");           
	printf("  -f		stay in the foreground for testing\n");
	printf("  -h		display this help\n");
	printf("  -i <address>	binds to interface <address>\n");
	printf("  -p <port>	sets network port (UDP and TCP)\n");
	printf("  -r <dir>	chroots to <dir>\n");
	printf("  -u <user>	switch to <user> (if started as root)\n");
	printf("  -V		display the version of this software\n");

	exit(1);
}

static void upsd_stop(int sig)
{
	ctype	*tmpcli;
	upstype	*ups, *unext;

	upslogx(LOG_INFO, "Signal %d: Shutting down", sig);

	/* cleanup client fds */
	for (tmpcli = firstclient; tmpcli != NULL; tmpcli = tmpcli->next) {
		shutdown(tmpcli->fd, shutdown_how);
		close(tmpcli->fd);
	}

	if (strcmp(pidfn, "") != 0)
		unlink(pidfn);

	ups = firstups;

	while (ups) {
		unext = ups->next;

		if (ups->sock_fd != -1)
			close(ups->sock_fd);

		sstate_infofree(ups);
		sstate_cmdfree(ups);
		pconf_finish(&ups->sock_ctx);

		if (ups->fn)
			free(ups->fn);
		if (ups->name)
			free(ups->name);
		free(ups);

		ups = unext;
	}

	/* dump everything */

	delete_acls();
	delete_access();
	user_flush();

	if (statepath)
		free(statepath);
	if (udpclient)
		free(udpclient);
	if (certfile)
		free(certfile);

	exit(0);
}

static void set_reload_flag(int sig)
{
	reload = 1;
}

/* basic signal setup to ignore SIGPIPE */
static void setupsignals(void)
{
	sa.sa_handler = SIG_IGN;
	sigemptyset (&upsd_sigmask);
	sa.sa_mask = upsd_sigmask;
	sa.sa_flags = 0;
	sigaction (SIGPIPE, &sa, NULL);

	/* handle shutdown signals */
	sa.sa_handler = upsd_stop;
	sigaction(SIGINT, &sa, NULL);
	sigaction(SIGQUIT, &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);

	/* handle reloading */
	sa.sa_handler = set_reload_flag;
	sigaction(SIGHUP, &sa, NULL);
}

int main(int argc, char **argv)
{
	int	i, cmd = 0;
	const char *nut_statepath_env = getenv("NUT_STATEPATH");
	int	do_background = 1;
	const	char *user = NULL;
	char	*progname, *chroot_path = NULL;
	struct stat st;
	struct	passwd	*new_uid = NULL;

	progname = argv[0];

	/* pick up a default from configure --with-user */
	user = RUN_AS_USER;

	/* yes, xstrdup - the conf handlers call free on this later */
	statepath = xstrdup(STATEPATH);

	/* set up some things for later */

	listenaddr.s_addr = INADDR_ANY;
	snprintf(pidfn, sizeof(pidfn), "%s/upsd.pid", ALTPIDPATH);

	printf("Network UPS Tools upsd %s\n", UPS_VERSION);

	while ((i = getopt(argc, argv, "+hp:r:i:fu:Vc:D")) != EOF) {
		switch (i) {
			case 'h':
				help(progname);
				break;
			case 'p':
				tcpport = udpport = atoi(optarg);
				break;
			case 'i':
				if (!inet_aton(optarg, &listenaddr))
					fatal("Invalid IP address");
				break;
			case 'r':
				chroot_path = optarg;
				break;
			case 'f':
				do_background = 0;
				break;
			case 'u':
				user = optarg;
				break;
			case 'V':

				/* do nothing - we already printed the banner */
				exit(0);

			case 'c':
				if (!strncmp(optarg, "reload", strlen(optarg)))
					cmd = SIGCMD_RELOAD;
				if (!strncmp(optarg, "stop", strlen(optarg)))
					cmd = SIGCMD_STOP;

				/* bad command given */
				if (cmd == 0)
					help(progname);
				break;

			case 'D':
				do_background = 0;
				nut_debug_level++;
				break;
			default:
				help(progname);
				break;
		}
	}

	if (cmd) {
		sendsignalfn(pidfn, cmd);
		exit(0);
	}

	argc -= optind;
	argv += optind;

	if (argc != 0)
		help(progname);

	/*
	 * currclient is pointed at this when dealing with udp clients 
	 * so that sendback uses the udpfd instead of a specific tcp socket
	 */

	udpclient = xmalloc(sizeof(ctype));
	udpclient->ssl = NULL;			/* impossible */

	setupsignals();
	setupudp();
	setuptcp();

	open_syslog("upsd");

	/* send logging to the syslog pre-background for later use */
        xbit_set(&upslog_flags, UPSLOG_SYSLOG);

	if (nut_statepath_env)
		statepath = xstrdup(nut_statepath_env);

	/* do this here, since getpwnam() might not work in the chroot */
	if ((new_uid = get_user_pwent(user)) == NULL)
		fatal("getpwnam(%s)", user);

	if (chroot_path)
		chroot_start(chroot_path);

	become_user(new_uid);

	if (chdir(statepath))
		fatal("Can't chdir to %s", statepath);

	/* check statepath perms */
	if (stat(".", &st))
		fatal("stat .");
	if (st.st_mode & (S_IROTH | S_IXOTH))
		upslogx(LOG_WARNING, "%s is world readable", statepath);

	/* read in the ups definitions (required) */
	read_upsconf(1);

	/* ... then do something with them */
	upsconf_add(0);			/* 0 = initial */

	read_upsdconf(0);		/* 0 = initial */
	user_load();

	if (upsnum == 0)
		fatalx("Fatal error: you need to define at least one UPS in your ups.conf!\n");

	ssl_init();

	/* this is after the uid change to detect permission problems */
	check_every_ups();

	if (do_background) {
		background();

		writepid(pidfn);
	} else {
		memset(&pidfn, '\0', sizeof(pidfn));
	}

	for (;;)
		mainloop();

	/* NOTREACHED */
	return 0;
}
